/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * Signal.hpp
 *
 *  Created on: 2010/11/08
 *      Author: HARK Development Team
 */

#ifndef SIGNAL_HPP_
#define SIGNAL_HPP_

#include <complex>
#include <valarray>
#include <stdexcept>
#include <boost/shared_ptr.hpp>

#include <Logger.hpp>

#include <Matrix.h>
#include <Vector.h>

namespace NSHarkBinaural
{

/**
 * 一般的な信号を表す.
 */
template <typename T>
class SignalBase
{
public:
    /**
     * 信号の要素の型
     */
    typedef T ElementType;

    /**
     * スマートポインタ型
     */
    typedef typename boost::shared_ptr<SignalBase> ptr_t;

    /**
     * 各要素が値valueで初期化された、要素数がsizeの信号を作成する.
     * @param size 信号の要素数
     * @param value 初期値
     */
    SignalBase(size_t size, const T& value = T())
        : data(T(), size), logger(Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME))
    {
        //DebugPuts("SignalBase::SignalBase(size_t, const T&)");
        logger.Finest(boost::format("SignalBase::SignalBase(size_t, const T&) %08x") %
                      reinterpret_cast<unsigned long>(this));
    }

    /**
     * 各要素が配列ptrの対応する要素で初期化された、要素数がsizeの信号を作成する.
     * @param size 信号の要素数
     * @param ptr 初期値が格納されている配列
     */
    SignalBase(size_t size, const T* ptr)
        : data(ptr, size), logger(Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME))
    {
        //DebugPuts("SignalBase::SignalBase(size_t, const T*)");
        logger.Finest(boost::format("SignalBase::SignalBase(size_t, const T*) %08x") %
                      reinterpret_cast<unsigned long>(this));
    }

    SignalBase(const SignalBase& rhs)
        : data(rhs.data), logger(Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME))
    {
        logger.Finest(boost::format("SignalBase::SignalBase(const SignalBase&) %08x") %
                      reinterpret_cast<unsigned long>(this));
    }

    /**
     * インスタンスを破棄する.
     */
    ~SignalBase()
    {
        //DebugPuts("SignalBase::~SignalBase()");
        logger.Finest(boost::format("SignalBase::~SignalBase() %08x") %
                      reinterpret_cast<unsigned long>(this));
    }

#if 0
    /**
     * 中身の信号配列を取得する.
     * @return 信号配列
     */
    std::valarray<T>& Data()
    {
        return this->data;
    }

    /**
     * 中身の信号配列を取得する.
     * @return 信号配列
     */
    const std::valarray<T>& Data() const
    {
        return this->data;
    }
#endif
    // valarrayへの参照を返すのではなく、valarrayの各メソッドを呼び出すような設計にする。
    // そうすれば、さまざまな変更に対して柔軟になる。
    // 例えば、「valarrayの各要素アクセスのログを取りたい」「遅延評価方式にしたい」とか。

    /**
     * 指定されたインデックスにある要素を得る.
     * @param index 要素のインデックス
     * @return 要素への参照
     */
    T& operator [](size_t index)
    {
        return this->data[index];
    }

    /**
     * 指定されたインデックスにある要素を得る.
     * @param index 要素のインデックス
     * @return 要素への参照
     */
    const T& operator [](size_t index) const
    {
        return this->data[index];
    }

    /**
     * 指定されたインデックスにある要素を得る.
     * 指定されたインデックスにある要素を得る。
     * もしインデックスが範囲外ならstd::out_of_range例外を送出する。
     * @param index 要素のインデックス
     * @return 要素への参照
     */
    T& At(size_t index)
    {
        CheckRange(index);
        return this->data[index];
    }

    /**
     * 指定されたインデックスにある要素を得る.
     * 指定されたインデックスにある要素を得る。
     * もしインデックスが範囲外ならstd::out_of_range例外を送出する。
     * @param index 要素のインデックス
     * @return 要素への参照
     */
    const T& At(size_t index) const
    {
        CheckRange(index);
        return this->data[index];
    }

    /**
     * 配列の要素数を得る.
     * @return 要素数
     */
    size_t Size() const
    {
        return this->data.size();
    }

    /**
     * 配列を、指定された初期値を持つ指定された大きさの配列に変える.
     * @param size 新しい配列の大きさ
     * @param val 配列の初期値
     */
    void Resize(size_t size, T val = T())
    {
        this->data.resize(size, val);
    }

    /**
     * 引数とそれぞれの要素の積を自分自身へ代入する.
     */
    SignalBase& operator *=(const T& rhs)
    {
        this->data *= rhs;
        return *this;
    }

    /**
     * 引数とそれぞれの要素の商を自分自身へ代入する.
     */
    SignalBase& operator /=(const T& rhs)
    {
        this->data /= rhs;
        return *this;
    }

    /**
     * 引数とそれぞれの要素の除算の余りを自分自身へ代入する.
     */
    SignalBase& operator %=(const T& rhs)
    {
        this->data %= rhs;
        return *this;
    }

    /**
     * 引数とそれぞれの要素の加算を自分自身へ代入する.
     */
    SignalBase& operator +=(const T& rhs)
    {
        this->data %= rhs;
        return *this;
    }

    /**
     * 引数とそれぞれの要素の減算を自分自身へ代入する.
     */
    SignalBase& operator -=(const T& rhs)
    {
        this->data %= rhs;
        return *this;
    }

    /**
     * 引数とそれぞれの要素のXORを自分自身へ代入する.
     */
    SignalBase& operator ^=(const T& rhs)
    {
        this->data ^= rhs;
        return *this;
    }

    /**
     * 引数とそれぞれの要素のANDを自分自身へ代入する.
     */
    SignalBase& operator &=(const T& rhs)
    {
        this->data &= rhs;
        return *this;
    }

    /**
     * 引数とそれぞれの要素のORを自分自身へ代入する.
     */
    SignalBase& operator |=(const T& rhs)
    {
        this->data %= rhs;
        return *this;
    }

    /**
     * 引数とそれぞれの要素の左シフトを自分自身へ代入する.
     */
    SignalBase& operator <<=(const T& rhs)
    {
        this->data <<= rhs;
        return *this;
    }

    /**
     * 引数とそれぞれの要素の右シフトを自分自身へ代入する.
     */
    SignalBase& operator >>=(const T& rhs)
    {
        this->data >>= rhs;
        return *this;
    }

private:
    /**
     * 信号を格納する配列
     */
    std::valarray<ElementType> data;

    Logger& logger;

    /**
     * 添字の範囲をチェックし、範囲外ならout_of_range例外を送出する.
     * @param index 検査する添字
     */
    void CheckRange(size_t index) const
    {
        if (index >= this->data.size())
        {
            throw std::out_of_range("index is out of range");
        }
    }
};

/**
 * 標準的に使う信号の要素の型.
 * 標準的に使う信号の要素の型を表す。
 * HARKやflowdesignerで使う型と一致させると、最適化されやすい。
 */
typedef float NormalSignalType;

/**
 * 時間領域信号を表す.
 */
typedef SignalBase<NormalSignalType> Unspectrum;

/**
 * 周波数領域信号を表す.
 */
typedef SignalBase<std::complex<NormalSignalType> > Spectrum;

/**
 * ToSignalArray()の実装.
 * 変換元の行列の要素と変換先の配列の要素の型が違う場合の実装
 */
template <typename T, typename U>
struct ToSignalArrayImpl
{
    static boost::shared_ptr<std::vector<typename SignalBase<T>::ptr_t> > ToSignalArray(const FD::Matrix<U>& matrix)
    {
        typedef boost::shared_ptr<std::vector<typename SignalBase<T>::ptr_t> > SignalArrayPtr;

        SignalArrayPtr result(new std::vector<typename SignalBase<T>::ptr_t>(matrix.nrows()));

        typename SignalBase<T>::ptr_t signalPtr;
        const U* matrixRowPtr = 0;

        for (size_t row = 0; row < matrix.nrows(); ++row)
        {
            matrixRowPtr = matrix[row];
            signalPtr = typename SignalBase<T>::ptr_t(new SignalBase<T>(matrix.ncols()));
            for (size_t col = 0; col < matrix.ncols(); ++col)
            {
                signalPtr->Data()[col] = static_cast<T>(matrixRowPtr[col]);
            }
            result->at(row) = signalPtr;
        }

        return result;
    }
};

/**
 * ToSignalArray()の実装.
 * 変換元の行列の要素と変換先の配列の要素の型が同じ場合に最適化した実装
 */
template <typename T>
struct ToSignalArrayImpl<T, T>
{
    static boost::shared_ptr<std::vector<typename SignalBase<T>::ptr_t> > ToSignalArray(const FD::Matrix<T>& matrix)
    {
        typedef boost::shared_ptr<std::vector<typename SignalBase<T>::ptr_t> > SignalArrayPtr;

        SignalArrayPtr result(new std::vector<typename SignalBase<T>::ptr_t>(matrix.nrows()));

        const T* matrixRowPtr = 0;

        for (size_t row = 0; row < matrix.nrows(); ++row)
        {
            matrixRowPtr = matrix[row];
            result->at(row) = typename SignalBase<T>::ptr_t(new SignalBase<T>(matrix.ncols(), matrixRowPtr));
        }

        return result;
    }
};

/**
 * FD::MatrixからSignalの配列へ変換する.
 * Signalの配列のインデックスと、FD::Matrixの行番号が対応するように変換される。
 * 変換結果をresultとすると、
 * result->at(i) と matrix[i] が対応する。
 * @param matrix 変換元の行列
 */
template <typename T, typename U>
boost::shared_ptr<std::vector<typename SignalBase<T>::ptr_t> > ToSignalArray(const FD::Matrix<U>& matrix)
{
    return ToSignalArrayImpl<T, U>::ToSignalArray(matrix);
}

template <typename T>
FD::RCPtr<FD::Vector<T> > ToFDVector(boost::shared_ptr<SignalBase<T> > signal)
{
    FD::RCPtr<FD::Vector<T> > result;
    if (signal.get() != 0)
    {
        result = FD::RCPtr<FD::Vector<T> >(new FD::Vector<T>(signal->Size()));
        for (size_t i = 0; i < signal->Size(); ++i)
        {
            result->operator [](i) = signal->operator [](i);
        }
    }
    return result;
}

/**
 * 対応する時間領域信号と周波数領域信号の組を格納する.
 * 時間領域信号と、それをフーリエ変換した周波数領域信号は、
 * 本質的には同一の信号である。
 * このクラスは、それら2つの信号の組を表している。
 */
class SignalHolder
{
public:
    typedef boost::shared_ptr<Unspectrum> UnspectrumPointer;
    typedef boost::shared_ptr<Spectrum> SpectrumPointer;

    /**
     * スマートポインタ型
     */
    typedef boost::shared_ptr<SignalHolder> ptr_t;

    /**
     * 時間領域信号と周波数領域信号の組を生成する.
     */
    SignalHolder()
        : unspectrum(UnspectrumPointer()), spectrum(SpectrumPointer())
    { }

    /**
     * 時間領域信号と周波数領域信号の組を生成する.
     * @param unspectrum 時間領域信号
     */
    SignalHolder(UnspectrumPointer unspectrum)
        : unspectrum(unspectrum), spectrum(SpectrumPointer())
    { }

    /**
     * 時間領域信号と周波数領域信号の組を生成する.
     * @param spectrum 周波数領域信号
     */
    SignalHolder(SpectrumPointer spectrum)
        : unspectrum(UnspectrumPointer()), spectrum(spectrum)
    { }

    /**
     * 時間領域信号と周波数領域信号の組を生成する.
     * @param unspectrum 時間領域信号
     * @param spectrum 周波数領域信号
     */
    SignalHolder(UnspectrumPointer unspectrum, SpectrumPointer spectrum)
        : unspectrum(unspectrum), spectrum(spectrum)
    { }

    /**
     * 時間領域信号を取得する.
     * @return 時間領域信号
     */
    UnspectrumPointer& GetUnspectrum()
    {
        return this->unspectrum;
    }

    /**
     * 時間領域信号を取得する.
     * @return 時間領域信号
     */
    const UnspectrumPointer& GetUnspectrum() const
    {
        return this->unspectrum;
    }

    /**
     * 周波数領域信号を取得する.
     * @return 周波数領域信号
     */
    SpectrumPointer& GetSpectrum()
    {
        return this->spectrum;
    }

    /**
     * 周波数領域信号を取得する.
     * @return 周波数領域信号
     */
    const SpectrumPointer& GetSpectrum() const
    {
        return this->spectrum;
    }

    /**
     * 時間領域信号を格納する.
     * @param unspectrum 時間領域信号
     */
    void SetUnspectrum(UnspectrumPointer unspectrum)
    {
        this->unspectrum = unspectrum;
    }

    /**
     * 周波数領域信号を格納する.
     * @param spectrum 周波数領域信号
     */
    void SetSpectrum(SpectrumPointer spectrum)
    {
        this->spectrum = spectrum;
    }

private:

    /**
     * 周波数領域信号に対応する時間領域信号.
     * 周波数領域信号しかない場合、このポインタは何も指さない
     */
    UnspectrumPointer unspectrum;

    /**
     * 時間領域信号に対応する周波数領域信号.
     * 時間領域信号しかない場合、このポインタは何も指さない
     */
    SpectrumPointer spectrum;

};

}

#endif /* SIGNAL_HPP_ */
